package bushfire.estimator;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import bushfire.model.Model;

/**
 * An estimator which is a composite of others. It simply applies all the other
 * estimators, then averages their values.
 * 
 * TODO: Turn this into particle filter of its own, with weights on the estimators,
 * adjusted according to how well they predict observations.
 *  
 * @author malcolmr
 */
public class CompositeEstimator extends Estimator {

	private Set myEstimators;
	private Map myWeights;

	/**
	 * @param model
	 */
	public CompositeEstimator(Model model) {
		super(model);

		myEstimators = new HashSet();
		myWeights = new HashMap();
	}

	public void addEstimator(Estimator estimator) {
		addEstimator(estimator, 1.0);
	}

	public void addEstimator(Estimator estimator, double weight) {
		myEstimators.add(estimator);
		myWeights.put(estimator, new Double(weight));
	}

	/**
	 * Return the average probability from all the estimators
	 * 
	 * @see bushfire.estimator.Estimator#updateStoredProbs()
	 */
	protected void updateStoredProbs() {

		clearDirtyFlag();

		int width = myModel.getWidth();
		int height = myModel.getHeight();

		double totalWeight = 0.0;

		for (Iterator it = myEstimators.iterator(); it.hasNext();) {
			Estimator e = (Estimator) it.next();

			e.updateStoredProbs();

			Double weight = (Double) myWeights.get(e);
			totalWeight += weight.doubleValue();
		}

		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				myEstimatedProbs[x][y][Model.UNBURNT] = 0.0;
				myEstimatedProbs[x][y][Model.BURNING] = 0.0;
				myEstimatedProbs[x][y][Model.BURNT] = 0.0;

				mySupport[x][y][Model.UNBURNT] = 0;
				mySupport[x][y][Model.BURNING] = 0;
				mySupport[x][y][Model.BURNT] = 0;

				for (Iterator it = myEstimators.iterator(); it.hasNext();) {
					Estimator e = (Estimator) it.next();
					Double weight = (Double) myWeights.get(e);

					myEstimatedProbs[x][y][Model.UNBURNT]
						+= e.getProb(x, y, Model.UNBURNT)
						* weight.doubleValue();
					myEstimatedProbs[x][y][Model.BURNING]
						+= e.getProb(x, y, Model.BURNING)
						* weight.doubleValue();
					myEstimatedProbs[x][y][Model.BURNT]
						+= e.getProb(x, y, Model.BURNT)
						* weight.doubleValue();

					mySupport[x][y][Model.UNBURNT]
						+= e.getSupport(x, y, Model.UNBURNT);
					mySupport[x][y][Model.BURNING]
						+= e.getSupport(x, y, Model.BURNING);
					mySupport[x][y][Model.BURNT]
						+= e.getSupport(x, y, Model.BURNT);
				}

				myEstimatedProbs[x][y][Model.UNBURNT] /= totalWeight;
				myEstimatedProbs[x][y][Model.BURNING] /= totalWeight;
				myEstimatedProbs[x][y][Model.BURNT] /= totalWeight;
			}
		}

	}

	/**
	 * @see bushfire.estimator.Estimator#nextStep()
	 */
	protected void nextStep() {
		// No need for the below: all the estimators observe the model.
		
//		for (Iterator it = myEstimators.iterator(); it.hasNext();) {
//			Estimator e = (Estimator) it.next();
//
//			e.nextStep();
//		}
		setDirtyFlag();
	}

	/**
	 * @see bushfire.estimator.Estimator#observe(int, int)
	 */
	public void observe(int x, int y) {
		for (Iterator it = myEstimators.iterator(); it.hasNext();) {
			Estimator e = (Estimator) it.next();

			e.observe(x, y);
		}
		setDirtyFlag();

	}

	/**
	 * @see bushfire.estimator.Estimator#observe(java.util.Set)
	 */
	public void observe(Set observations) {
		for (Iterator it = myEstimators.iterator(); it.hasNext();) {
			Estimator e = (Estimator) it.next();

			e.observe(observations);
		}

		setDirtyFlag();
	}

	/**
	 * @see bushfire.estimator.Estimator#reset()
	 */
	protected void reset() {
		// No need for the below: all the estimators observe the model.
//		for (Iterator it = myEstimators.iterator(); it.hasNext();) {
//			Estimator e = (Estimator) it.next();
//
//			e.reset();
//		}
		setDirtyFlag();
	}

	/**
	 * @see bushfire.estimator.Estimator#setFireAt(int, int, byte)
	 */
	protected void setFireAt(int x, int y, byte state) {
		// No need for the below: all the estimators observe the model.
//		for (Iterator it = myEstimators.iterator(); it.hasNext();) {
//			Estimator e = (Estimator) it.next();
//
//			e.setFireAt(x, y, state);
//		}

		setDirtyFlag();
	}

	public String toString() {
		return "Composite Estimator";
	}

}
