package bushfire.estimator;

import java.util.Observable;
import java.util.Observer;
import java.util.Set;

import bushfire.model.Model;

/**
 * This class represents an abstract estimator
 * 
 * It observes the Model and updates itself in accordance with changes,
 * due to resetting, user-specified changes, or transition steps.
 * 
 * It is also observable, so UI components can be notified when it changes. 
 * 
 * @author malcolmr
 */
abstract public class Estimator extends Observable implements Observer {

	private boolean myDirtyFlag;

	protected Model myModel;

	protected double[][][] myEstimatedProbs;
	protected int[][][] mySupport;

	public Estimator(Model model) {
		myModel = model;
		myModel.addObserver(this);

		myEstimatedProbs = new double[model.getWidth()][model.getHeight()][3];
		mySupport = new int[model.getWidth()][model.getHeight()][3];
		setDirtyFlag();
	}

	protected synchronized void clearDirtyFlag() {

		myDirtyFlag = false;
	}


	protected synchronized void setDirtyFlag() {

		myDirtyFlag = true;
		setChanged();
		notifyObservers();

	}

	public double getProb(int x, int y, int type) {
		if (myDirtyFlag) {
			updateStoredProbs();
		}
		return myEstimatedProbs[x][y][type];
	}

	public int getSupport(int x, int y, int type) {
		if (myDirtyFlag) {
			updateStoredProbs();
		}
		return mySupport[x][y][type];
	}

	abstract protected void updateStoredProbs();
	
	abstract protected void nextStep();

	/**
	 * Make a single observation.
	 * @param x
	 * @param y
	 */

	abstract public void observe(int x, int y);
	
	/**
	 * Make a bunch of observations, saving resampling until the end
	 * @param myScheduledObservations
	 */
	abstract public void observe(Set observations);
	
	
	/**
	 * @return
	 */
	public int getWidth() {
		return myModel.getWidth();
	}

	public int getHeight() {
		return myModel.getHeight();
	}


	/**
	 * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
	 */
	public void update(Observable o, Object arg) {
		if (arg == null) {
			nextStep();
		} else if (arg instanceof Model.FirePoint) {
			Model.FirePoint p = (Model.FirePoint) arg;
			setFireAt(p.getX(), p.getY(), p.getState());
		} else {
			reset();
		}
	}

	/**
	 * 
	 */
	abstract protected void reset();
	
	/**
	 * @param d
	 * @param e
	 * @param i
	 */
	abstract protected void setFireAt(int x, int y, byte state);
	
}
