package bushfire.estimator;

import java.awt.Point;
import java.util.Iterator;
import java.util.Set;

import bushfire.model.Model;

/**
 * This class represents a particle filer.
 * 
 * @author malcolmr
 */
public class MonolithicEstimator extends Estimator {

	private static final int DEFAULT_NPARTICLES = 1000;

	private int myNParticles;
	private Particle[] myParticles;

	public MonolithicEstimator(Model model) {
		this(model, DEFAULT_NPARTICLES);
	}

	public MonolithicEstimator(Model model, int nParticles) {
		super(model);
		myNParticles = nParticles;
		myParticles = new Particle[myNParticles];

		for (int i = 0; i < myNParticles; i++) {
			myParticles[i] = new Particle(model);
		}
	}

	protected synchronized void updateStoredProbs() {

		clearDirtyFlag();
		int width = myModel.getWidth();
		int height = myModel.getHeight();

		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				myEstimatedProbs[x][y][Model.UNBURNT] = 0.0;
				myEstimatedProbs[x][y][Model.BURNING] = 0.0;
				myEstimatedProbs[x][y][Model.BURNT] = 0.0;

				mySupport[x][y][Model.UNBURNT] = 0;
				mySupport[x][y][Model.BURNING] = 0;
				mySupport[x][y][Model.BURNT] = 0;
			}
		}

		double totalWeight = 0.0;

		for (int i = 0; i < myNParticles; i++) {
			double weight = myParticles[i].getWeight();
			totalWeight += weight;
			for (int x = 0; x < width; x++) {
				for (int y = 0; y < height; y++) {
					int state = myParticles[i].getFireAt(x, y);
					myEstimatedProbs[x][y][state] += weight;
					mySupport[x][y][state]++;
				}
			}
		}

		// normalise

		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				myEstimatedProbs[x][y][Model.UNBURNT] /= totalWeight;
				myEstimatedProbs[x][y][Model.BURNING] /= totalWeight;
				myEstimatedProbs[x][y][Model.BURNT] /= totalWeight;
			}
		}

	}

	protected void nextStep() {
		for (int i = 0; i < myNParticles; i++) {
			myParticles[i].nextStep();
		}
		setDirtyFlag();
	}

	public void observe(int x, int y) {
		adjustWeights(x, y);
		renormalise();
		possiblyResample();

		setDirtyFlag();
	}

	/**
	 * Make a bunch of observations, saving resampling until the end
	 * @param myScheduledObservations
	 */
	public void observe(Set observations) {

		for (Iterator it = observations.iterator(); it.hasNext();) {
			Point p = (Point) it.next();
			adjustWeights(p.x, p.y);
		}

		renormalise();
		possiblyResample();

		setDirtyFlag();
	}

	private void adjustWeights(int x, int y) {
		byte observation = myModel.observation(x, y);

		for (int i = 0; i < myNParticles; i++) {
			double weight = myParticles[i].getWeight();
			weight
				*= myParticles[i].observationProbability(x, y, observation);
			myParticles[i].setWeight(weight);
		}

	}

	private void renormalise() {
		double totalWeight = 0.0;
		
		for (int i = 0; i < myNParticles; i++) {
			totalWeight += myParticles[i].getWeight();
		}

		for (int i = 0; i < myNParticles; i++) {
			double weight = myParticles[i].getWeight();
			weight /= totalWeight;
			myParticles[i].setWeight(weight);
		}

	}

	private void possiblyResample() {

		double totalWeightSq = 0.0;
		for (int i = 0; i < myNParticles; i++) {
			double weight = myParticles[i].getWeight();
			totalWeightSq += weight * weight;
		}

		int effectiveParticles = (int) Math.round(1.0 / totalWeightSq);

		System.out.println("# effective particles = " + effectiveParticles);

		if (effectiveParticles < myNParticles / 2) {
			resampleParticles();
		}
	}

	/**
	 * 
	 */
	public void resampleParticles() {
		resampleParticles(myNParticles);
	}

	/**
	 * 
	 */
	public void resampleParticles(int nParticles) {
		Particle[] newParticles = new Particle[nParticles];

		int j = 0;

		double totalWeight = 0.0;
		for (int i = 0; i < myNParticles; i++) {
			totalWeight += myParticles[i].getWeight();
		}

		double delta = totalWeight / nParticles;
		double weight = 0.0;
		totalWeight = 0.0;

		for (int i = 0; i < myNParticles; i++) {
			totalWeight += myParticles[i].getWeight();;

			if (weight < totalWeight && j < nParticles) {
				// don't copy the first one, to save unnecessary garbage collection
				newParticles[j++] = myParticles[i];

				weight += delta;

				// subsequent duplicates are copies

				while (weight < totalWeight && j < nParticles) {
					newParticles[j++] = new Particle(myParticles[i]);
					weight += delta;
				}
			}
		}

		System.out.println("#particles = " + j);
		if (j != nParticles) {
			throw new RuntimeException("Not enough particles!");
		}

		myNParticles = nParticles;
		myParticles = newParticles;

		for (int i = 0; i < myNParticles; i++) {
			myParticles[i].setWeight(1.0);
		}
		setDirtyFlag();
	}

	/**
	 * @return
	 */
	public int getNParticles() {
		return myNParticles;
	}

	/**
	 * @param i
	 */
	public void setNParticles(int n) {
		resampleParticles(n);
	}

	/**
	 * 
	 */
	protected void reset() {
		for (int i = 0; i < myNParticles; i++) {
			myParticles[i] = new Particle(myModel);
		}

		hasChanged();
		notifyObservers();
		setDirtyFlag();
	}

	/**
	 * @param d
	 * @param e
	 * @param i
	 */
	protected void setFireAt(int x, int y, byte state) {
		for (int i = 0; i < myNParticles; i++) {
			myParticles[i].setFireAt(x, y, state);
		}

		setDirtyFlag();
		hasChanged();
		notifyObservers();
	}

	public String toString() {
		return "Monolithic Estimator";
	}

}
