package bushfire.estimator;

import java.awt.Dimension;
import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import bushfire.model.Model;

/**
 * This code implements a particle filter with many semi-independent "tiles",
 * corresponding to rectangles in the map. The assumption is that correlations 
 * are only local, so an observation only effects the estimated values in cells
 * local to the point it was taken - i.e. in its own tile. 
 * 
 * So, for example a 100x100 grid can be divided up into 100 10x10 tiles. Each
 * particle in the filter stored 10x10 different weights, one for each tile.
 * These weights are used in computing the probabilities of values of the cells 
 * in that tile.
 * 
 * Each particle has a since Model, as before, which is used to compute the 
 * transitions over all the tiles.
 * 
 * Observations only affect the weight of the tile in which the observation was
 * made. Other tiles in the particle are unaffected.
 * 
 * Resampling duplicates/discards tiles, not entire particles. Since the total
 * number of particles is constant, for each tile discarded, another is duplicated.
 * So the cells are copied from the duplicated tile into the discarded tile. The
 * rest of the affected particles are unchanged. 
 * 
 * NOTE: Using this estimator on its own cause strange discontinuities along the
 * edges of tiles. This can be overcome, however, by using a CompositeEstimator 
 * containing several TiledEstimators with different offsets or different cell 
 * sizes.
 *  
 * @author malcolmr
 */
public class TiledEstimator extends Estimator {

	private Dimension myDimension;
	private Point myOffset;
	
	private int myTileWidth;
	private int myTileHeight;
	private TiledParticle[] myParticles;
	private int myNParticles;

	/**
	 * Divide model into width by height tiles.
	 * 
	 * @param model
	 */
	public TiledEstimator(
		Model model,
		int nParticles,
		int tileWidth,
		int tileHeight) {

		this(model, nParticles, tileWidth, tileHeight, 0, 0);
	}

	/**
	 * Create a tiled estimator with the given offset.
	 * 
	 * @param model
	 * @param nParticles
	 * @param nAcross
	 * @param nDown
	 * @param xOffset
	 * @param yOffset
	 */

	public TiledEstimator(
		Model model,
		int nParticles,
		int tileWidth,
		int tileHeight,
		int xOffset,
		int yOffset) {
				
		super(model);
		myOffset = new Point(xOffset, yOffset);

		xOffset %= tileWidth;
		if (xOffset > 0) {
			xOffset -= tileWidth;
		}
		
		yOffset %= tileHeight;
		if (yOffset > 0) {
			yOffset -= tileHeight;
		}

		myOffset = new Point(xOffset, yOffset);
				
		myTileWidth = tileWidth;
		myTileHeight = tileHeight;

		int nAcross = 1 + (model.getWidth() - xOffset - 1) / tileWidth; 
		int nDown = 1 + (model.getHeight() - yOffset - 1) / tileHeight; 
		myDimension = new Dimension(nAcross, nDown);

		myNParticles = nParticles;
		myParticles = new TiledParticle[myNParticles];

		for (int i = 0; i < myNParticles; i++) {
			myParticles[i] = new TiledParticle(model, myDimension, myOffset);
		}
	}

	/**
	 * @see bushfire.estimator.Estimator#updateProbs()
	 */
	protected void updateStoredProbs() {
		clearDirtyFlag();
		int width = myModel.getWidth();
		int height = myModel.getHeight();

		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				Point tile = getTile(x, y);

				myEstimatedProbs[x][y][Model.UNBURNT] = 0.0;
				myEstimatedProbs[x][y][Model.BURNING] = 0.0;
				myEstimatedProbs[x][y][Model.BURNT] = 0.0;

				mySupport[x][y][Model.UNBURNT] = 0;
				mySupport[x][y][Model.BURNING] = 0;
				mySupport[x][y][Model.BURNT] = 0;

				double totalWeight = 0.0;

				for (int i = 0; i < myNParticles; i++) {
					double weight = myParticles[i].getTileWeight(tile);
					totalWeight += weight;

					int state = myParticles[i].getFireAt(x, y);
					myEstimatedProbs[x][y][state] += weight;
					mySupport[x][y][state]++;
				}

				myEstimatedProbs[x][y][Model.UNBURNT] /= totalWeight;
				myEstimatedProbs[x][y][Model.BURNING] /= totalWeight;
				myEstimatedProbs[x][y][Model.BURNT] /= totalWeight;
			}
		}
	}

	/**
	 * @see bushfire.estimator.Estimator#nextStep()
	 */
	protected void nextStep() {
		for (int i = 0; i < myNParticles; i++) {
			myParticles[i].nextStep();
		}
		setDirtyFlag();

	}

	/**
	 * @param x
	 * @param y
	 */
	private Point getTile(int x, int y) {
		Point tile = new Point();
		tile.x = (x - myOffset.x) / myTileWidth;
		tile.y = (y - myOffset.y) / myTileHeight;

		return tile;
	}

	/**
	 * @see bushfire.estimator.Estimator#observe(int, int)
	 */
	public void observe(int x, int y) {
		Point tile = getTile(x, y);

		adjustWeights(new Point(x, y), tile);
		renormalise(tile);
		possiblyResample(tile);

		setDirtyFlag();
	}

	/**
	 * @see bushfire.estimator.Estimator#observe(java.util.Set)
	 */
	public void observe(Set observations) {

		Set affectedTiles = new HashSet();

		for (Iterator it = observations.iterator(); it.hasNext();) {
			Point p = (Point) it.next();
			Point tile = getTile(p.x, p.y);

			adjustWeights(p, tile);
			affectedTiles.add(tile);
		}

		for (Iterator it = affectedTiles.iterator(); it.hasNext();) {
			Point tile = (Point) it.next();

			renormalise(tile);
			possiblyResample(tile);
		}

		setDirtyFlag();

	}

	/**
	 * @param tile
	 */
	private void adjustWeights(Point point, Point tile) {
		byte observation = myModel.observation(point.x, point.y);

		for (int i = 0; i < myNParticles; i++) {
			double weight = myParticles[i].getTileWeight(tile);
			weight
				*= myParticles[i].observationProbability(
					point.x,
					point.y,
					observation);
			myParticles[i].setTileWeight(tile, weight);
		}

	}

	private void renormalise(Point tile) {
		double totalWeight = 0.0;

		for (int i = 0; i < myNParticles; i++) {
			totalWeight += myParticles[i].getTileWeight(tile);
		}

		for (int i = 0; i < myNParticles; i++) {
			double weight = myParticles[i].getTileWeight(tile);
			weight /= totalWeight;
			myParticles[i].setTileWeight(tile, weight);
		}

	}

	private void possiblyResample(Point tile) {

		double totalWeightSq = 0.0;

		for (int i = 0; i < myNParticles; i++) {
			double weight = myParticles[i].getTileWeight(tile);
			totalWeightSq += weight * weight;
		}

		int effectiveParticles = (int) Math.round(1.0 / totalWeightSq);

		System.out.println("# effective particles = " + effectiveParticles);

		if (effectiveParticles < myNParticles / 2) {
			resampleTile(tile);
		}
	}

	private void resampleTile(Point tile) {

		System.out.println("Resampling tile: (" + tile.x + ", " + tile.y + ")");

		double totalWeight = 0.0;
		for (int i = 0; i < myNParticles; i++) {
			totalWeight += myParticles[i].getTileWeight(tile);
		}

		double delta = totalWeight / myNParticles;
		double weight = 0.0;
		totalWeight = 0.0;

		List discarded = new ArrayList();
		List duplicated = new ArrayList();

		int nKeptParticles = 0;

		for (int i = 0; i < myNParticles; i++) {
			totalWeight += myParticles[i].getTileWeight(tile);

			if (weight < totalWeight && nKeptParticles < myNParticles) {
				// just keep the first particle
				nKeptParticles++;
				weight += delta;

				// subsequent duplicates are copies

				while (weight < totalWeight && nKeptParticles < myNParticles) {
					nKeptParticles++;
					duplicated.add(myParticles[i]);
					weight += delta;
				}
			} else {
				discarded.add(myParticles[i]);
			}
		}

		if (duplicated.size() != discarded.size()) {
			throw new RuntimeException(
				"# Discarded Tiles "
					+ discarded.size()
					+ " != # duplicated tiles "
					+ duplicated.size());
		}

		// randomly assign duplicated particles to discarded particles
		Collections.shuffle(duplicated);

		Iterator it1 = duplicated.iterator();
		Iterator it2 = discarded.iterator();

		while (it1.hasNext()) {
			TiledParticle duplicatedParticle = (TiledParticle) it1.next();
			TiledParticle discardedParticle = (TiledParticle) it2.next();

			// Copy the tile from the duplicated particle to the discardedParticle

			discardedParticle.copyTile(duplicatedParticle, tile);
		}

		for (int i = 0; i < myNParticles; i++) {
			myParticles[i].setTileWeight(tile, 1.0);
		}
		setDirtyFlag();

	}

	/**
	 * @see bushfire.estimator.Estimator#reset()
	 */
	protected void reset() {
		for (int i = 0; i < myNParticles; i++) {
			myParticles[i] = new TiledParticle(myModel, myDimension, myOffset);
		}

		hasChanged();
		notifyObservers();
		setDirtyFlag();

	}

	/**
	 * @see bushfire.estimator.Estimator#setFireAt(int, int, byte)
	 */
	protected void setFireAt(int x, int y, byte state) {
		for (int i = 0; i < myNParticles; i++) {
			myParticles[i].setFireAt(x, y, state);
		}

		setDirtyFlag();
		hasChanged();
		notifyObservers();
	}

	public String toString() {
		return "TiledEstimator dim = " + myDimension + ", offset = " + myOffset;
	}

}
