package bushfire.model;

import java.util.HashMap;
import java.util.Iterator;

/**
 * An implementation of the abstract Model class, using a HashMap to map gridcells
 * to states. Unburnt cells are not represented in the map, empty cells are considered
 * unburnt by default.
 * 
 * Unfortunately the overhead of building the hashmap seems to eliminate any value
 * for reasonable large fires.
 * 
 * @author malcolmr
 */

public class HashModel extends Model {

	private HashMap myFireState;

	/**
	 * Create a new model of a fire, with the untire grid unburnt
	 * This method is protected, so Models must be created using Model.makeModel
	 * 
	 * @param width
	 * @param height
	 */

	protected HashModel(int width, int height) {
		super(width, height);

		myFireState = new HashMap();
		reset();
	}

	/**
	 * Set the entire grid back to unburnt. 
	 * 
	 */

	public void reset() {
		myFireState.clear();
		setChanged();
		notifyObservers(new ResetMessage());
	}

	/**
	 * Set the state of a particular cell to one of UNBURNT, BURNING or BURNT.
	 */

	public void setFireAt(int x, int y, byte state) {

		if (state != UNBURNT && state != BURNING && state != BURNT) {
			throw new IllegalArgumentException(
				state + " is not a valid fire state.");
		}

		if (state == UNBURNT) {
			myFireState.remove(new MyPoint2D(x, y));
		}
		else {
			myFireState.put(new MyPoint2D(x, y), new Byte(state));
		}

		setChanged();
		notifyObservers(new FirePoint(x, y, state));
	}

	/**
	 * Get the state of the fire at the given location
	 * @param x
	 * @param y
	 * @return
	 */

	public byte getFireAt(int x, int y) {
		Byte i = (Byte) myFireState.get(new MyPoint2D(x, y));
		return (i == null ? UNBURNT : i.byteValue());
	}

	public byte getFireAt(MyPoint2D p) {
		Byte i = (Byte) myFireState.get(p);
		return (i == null ? UNBURNT : i.byteValue());
	}

	/**
	 * Apply the transition model to all locations to get the state
	 * for the next timestep.
	 */

	public void nextStep() {

		HashMap nextState = new HashMap();

		for (Iterator it = myFireState.keySet().iterator(); it.hasNext();) {
			MyPoint2D p = (MyPoint2D) it.next();

			switch (getFireAt(p)) {

				case UNBURNT:
					// Nothing to do. 
					// Unburnt squares either remain unburnt (no entry)
					// Or else they are set on fire by a neighbouring BURNING
					// square. It's more efficient to do that computation below.
					break;					

				case BURNING :
					// Possibly die out

					if (Math.random() < DIE_PROB) {
						nextState.put(p, new Byte(BURNT));
					} else {
						nextState.put(p, new Byte(BURNING));
					}

					// Possibly set unburnt neighbours on fire

					for (int dx = -1; dx <= 1; dx++) {
						for (int dy = -1; dy <= 1; dy++) {
							if (p.myX + dx < 0
								|| p.myX + dx >= myWidth
								|| p.myY + dy < 0
								|| p.myY + dy >= myHeight) {

								continue;
							}

							MyPoint2D p1 =
								new MyPoint2D(p.myX + dx, p.myY + dy);
							if (getFireAt(p1) == UNBURNT
								&& Math.random() < CATCH_PROB) {
								nextState.put(p1, new Byte(BURNING));
							}
						}
					}
					break;

				case BURNT :
					nextState.put(p, new Byte(BURNT));
					break;
					
			}

		}
		myFireState = nextState;
		setChanged();
		notifyObservers(null);

	}
	class MyPoint2D {
		int myX;
		int myY;

		MyPoint2D(int x, int y) {
			myX = x;
			myY = y;
		}

		/**
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
		public boolean equals(Object obj) {
			MyPoint2D p = (MyPoint2D) obj;

			return (p.myX == myX) && (p.myY == myY);
		}

		/**
		 * @see java.lang.Object#hashCode()
		 */
		public int hashCode() {
			return myY * myWidth + myX;
		}

	}

}
