package bushfire.model;

import java.awt.Dimension;
import java.util.Observable;

/**
 * This class defines the state, transition and sensor models.
 * 
 * @author malcolmr
 */

abstract public class Model extends Observable {

	protected int myWidth;
	protected int myHeight;

	/**
	 * Note: We are assuming elsewhere that UNBURNT = 0
	 */

	public static final byte UNBURNT = 0;
	public static final byte BURNING = 1;
	public static final byte BURNT = 2;

	/**
	 * Static fields controlling the transition model for all instances
	 * of this class.
	 */

	protected static double CATCH_PROB = 0.15;
	protected static double DIE_PROB = 0.6;

	/**
	 * Factory method for making models, hiding the particular 
	 * implementing subclass.
	 * 
	 * @param width
	 * @param height
	 * @return
	 */

	static public Model makeModel(int width, int height) {
		return new SimpleModel(width, height);
	}

	/**
	 * Factory method for copying models, hiding the particular 
	 * implementing subclass.
	 * 
	 * @return
	 */

	static public Model copyModel(Model model) {

		int width = model.getWidth();
		int height = model.getHeight();

		Model m = makeModel(width, height);

		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				m.setFireAt(x, y, model.getFireAt(x, y));
			}

		}
		return m;
	}

	/**
	 * Create a new model of a fire, with the untire grid unburnt.
	 * This method is protected, so Models must be created using 
	 * the factory method above.
	 * 
	 * @param width
	 * @param height
	 */

	protected Model(int width, int height) {
		myWidth = width;
		myHeight = height;
	}
	
	/**
	 * Set the entire grid back to unburnt. 
	 * 
	 */

	public void reset() {
		for (int x = 0; x < myWidth; x++) {
			for (int y = 0; y < myHeight; y++) {
				setFireAt(x, y, UNBURNT);
			}
		}
		setChanged();
		notifyObservers(new ResetMessage());
	}

	/**
	 * Set the state of a particular cell to one of UNBURNT, BURNING or BURNT.
	 */

	abstract public void setFireAt(int x, int y, byte state);
	
	/**
	 * Get the state of the fire at the given location
	 * @param x
	 * @param y
	 * @return
	 */

	abstract public byte getFireAt(int x, int y);

	/**
	 * Apply the transition model to all locations to get the state
	 * for the next timestep.
	 */

	abstract public void nextStep();
	
	public static double getCatchProb() {
		return CATCH_PROB;
	}

	public static double getDieProb() {
		return DIE_PROB;
	}

	public static void setCatchProb(double prob) {
		CATCH_PROB = prob;
	}

	public static void setDieProb(double prob) {
		DIE_PROB = prob;
	}

	public int getHeight() {
		return myHeight;
	}

	public int getWidth() {
		return myWidth;
	}

	public Dimension getDimension() {
		return new Dimension(myWidth, myHeight);
	}

	/**
	 * Check if the given point is on the map
	 */
	public boolean onMap(int x, int y) {
		return x >= 0 && x < myWidth && y >= 0 && y < myHeight;
	}


	private static final double[][] OBSERVE_PROBS = { { 0.9, 0.05, 0.05 }, {
			0.05, 0.9, 0.05 }, {
			0.05, 0.05, 0.9 }

	};

	/**
	 * Make an observation
	 */

	public byte observation(int x, int y) {

		byte actual = getFireAt(x, y);
		return actual;
	}

	/**
	 * Find the probability that the given observation would have been 
	 * made a the given location.
	 * 
	 * @param x
	 * @param y
	 * @param observation
	 * @return
	 */

	public double observationProbability(int x, int y, byte observation) {

		// Blurred

		byte actual = getFireAt(x, y);
		return OBSERVE_PROBS[actual][observation];
	}
	
	/**
	 * Compute the distance between this model and another.
	 * Distance = # of cells non-matching.
	 * @param m
	 * @return
	 */

	public int distance(Model m) {
		if (myWidth != m.myWidth || myHeight != m.myHeight) {
			throw new RuntimeException("Attempted to compare models of unequal dimensions.");
		}
		
		int differences = 0;
		
		for (int x = 0; x < myWidth; x++) {
			for (int y = 0; y < myHeight; y++) {
				if (getFireAt(x,y) != m.getFireAt(x, y)) {
					differences++;
				}
			}
		}
		
		return differences;
	}

	static public class FirePoint {

		private int myX;
		private int myY;
		private byte myState;

		/**
		 * @param x
		 * @param y
		 */
		public FirePoint(int x, int y, byte state) {
			myX = x;
			myY = y;
			myState = state;
		}


		/**
		 * @return
		 */
		public byte getState() {
			return myState;
		}

		/**
		 * @return
		 */
		public int getX() {
			return myX;
		}

		/**
		 * @return
		 */
		public int getY() {
			return myY;
		}

	}

	static public class ResetMessage {

	}

}
