package bushfire.model;


/**
 * This class defines the state, transition and sensor models.
 * 
 * @author malcolmr
 */

public class SimpleModel extends Model {

	private byte[][] myFireState;

	/**
	 * Create a new model of a fire, with the untire grid unburnt.
	 * This method is protected, so Models must be created using Model.makeModel
	 * 
	 * @param width
	 * @param height
	 */

	protected SimpleModel(int width, int height) {
		super(width, height);

		myFireState = new byte[width][height];
		reset();
	}

	/**
	 * Set the entire grid back to unburnt. 
	 * 
	 */

	public void reset() {
		for (int x = 0; x < myWidth; x++) {
			for (int y = 0; y < myHeight; y++) {
				myFireState[x][y] = UNBURNT;
			}
		}
		setChanged();
		notifyObservers(new ResetMessage());
	}

	/**
	 * Set the state of a particular cell to one of UNBURNT, BURNING or BURNT.
	 */

	public void setFireAt(int x, int y, byte state) {

		if (state != UNBURNT && state != BURNING && state != BURNT) {
			throw new IllegalArgumentException(
				state + " is not a valid fire state.");
		}
		myFireState[x][y] = state;

		setChanged();
		notifyObservers(new FirePoint(x, y, state));
	}

	/**
	 * Get the state of the fire at the given location
	 * @param x
	 * @param y
	 * @return
	 */

	public byte getFireAt(int x, int y) {
		return myFireState[x][y];
	}

	/**
	 * Apply the transition model to all locations to get the state
	 * for the next timestep.
	 */

	public void nextStep() {

		byte[][] nextState = new byte[myWidth][myHeight];

		for (int x = 0; x < myWidth; x++) {
			for (int y = 0; y < myHeight; y++) {

				switch (myFireState[x][y]) {

					case UNBURNT :
						// Unburnt squares either remain unburnt (zero)
						// Or else they are set on fire by a neighbouring BURNING
						// square. It's more efficient to do that computation below.
						break;

					case BURNING :
						if (Math.random() < DIE_PROB) {
							nextState[x][y] = BURNT;
						} else {
							nextState[x][y] = BURNING;
						}
						// Possibly set unburnt neighbours on fire

						for (int dx = -1; dx <= 1; dx++) {
							for (int dy = -1; dy <= 1; dy++) {
								if (x + dx < 0
									|| x + dx >= myWidth
									|| y + dy < 0
									|| y + dy >= myHeight) {

									continue;
								}

								if (myFireState[x+dx][y+dy] == UNBURNT
									&& Math.random() < CATCH_PROB) {
									nextState[x+dx][y+dy] = BURNING;
								}
							}
						}
						break;

					case BURNT :
						nextState[x][y] = BURNT;
						break;
				}
			}
		}

		myFireState = nextState;
		setChanged();
		notifyObservers(null);

	}

}
