package bushfire.ui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Observable;
import java.util.Observer;
import java.util.Set;

import javax.swing.JComponent;
import javax.swing.JLabel;

import bushfire.estimator.Estimator;
import bushfire.model.Model;

/**
 * A graphical representation of the contents of the Estimator.
 * 
 * @author malcolmr
 */
public class EstimatorView extends JLabel implements Observer {

	private Estimator myEstimator;
	private static final int CELL_SIZE = 10;

	private Set myScheduledObservations;

	public EstimatorView(Estimator estimator) {
		estimator.addObserver(this);
		myEstimator = estimator;

		Dimension size = getPreferredSize();

		MyMouseListener listener = new MyMouseListener();
		addMouseListener(listener);
		addMouseMotionListener(listener);

		myScheduledObservations = new HashSet();

	}

	public void reset() {
		myScheduledObservations = new HashSet();
		repaint();
	}

	/**
	 * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
	 */
	public void update(Observable o, Object arg) {
		repaint();
	}

	/**
	 * @see java.awt.Component#getPreferredSize()
	 */
	public Dimension getPreferredSize() {
		return new Dimension(
			myEstimator.getWidth() * CELL_SIZE,
			myEstimator.getHeight() * CELL_SIZE);
	}

	public class MyMouseListener
		extends MouseAdapter
		implements MouseMotionListener {

		boolean myDragging = false;
		boolean myDragValue = true;

		/**
		* @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
		*/
		public void mouseClicked(MouseEvent e) {
			Point p = pointToCell(e.getPoint());
			
			System.out.println("mouseClicked");
			System.out.println("p = " + p);
			
			if (p != null) {
				scheduleObservation(p, !myScheduledObservations.contains(p));				
			}

		}

		/**
		 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
		 */
		public void mouseDragged(MouseEvent e) {
			Rectangle r = new Rectangle(e.getX(), e.getY(), 1, 1);
			((JComponent) e.getSource()).scrollRectToVisible(r);
			Point p = pointToCell(e.getPoint());

			if (p == null) {
				return;
			}
			
			if (!myDragging) {
				myDragging = true;
				myDragValue = !myScheduledObservations.contains(p);
			}

			scheduleObservation(p, myDragValue);
		}

		/**
		 * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
		 */
		public void mouseMoved(MouseEvent e) {
		}

		/**
		 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
		 */
		public void mouseReleased(MouseEvent e) {
			myDragging = false;
		}

	}

	private void scheduleObservation(Point p, boolean on) {

		if (on) {
			myScheduledObservations.add(p);
		}
		else {
			myScheduledObservations.remove(p);
		}
		
		repaint(p.x * CELL_SIZE, p.y * CELL_SIZE, CELL_SIZE, CELL_SIZE);
	}

	private void paintGrid(Graphics g) {
		int width = myEstimator.getWidth();
		int height = myEstimator.getHeight();

		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {

				double unburntProb = myEstimator.getProb(x, y, Model.UNBURNT);
				double burningProb = myEstimator.getProb(x, y, Model.BURNING);
				double burntProb = myEstimator.getProb(x, y, Model.BURNT);

				// Colour will be white if definitely unburnt,
				//                  red if definitely burning,
				//                black if definitely burnt
				// and somewhere in between for other variations

				g.setColor(
					Color.getHSBColor(
						(float) 0.0,
						(float) burningProb,
						(float) (1 - burntProb)));

				g.fillRect(x * CELL_SIZE, y * CELL_SIZE, CELL_SIZE, CELL_SIZE);
			}		
		}
		
		// highlight scheduled observations
		
		g.setColor(new Color((float)0.0, (float)0.0, (float)1.0, (float)0.5));

		for (Iterator it = myScheduledObservations.iterator(); it.hasNext();) {
			Point p = (Point) it.next();
			
			g.fillRect(p.x * CELL_SIZE, p.y * CELL_SIZE, CELL_SIZE, CELL_SIZE);			
		}

		// draw gridlines

		g.setColor(Color.GRAY);
		for (int x = 0; x <= width; x++) {
			g.drawLine(x * CELL_SIZE, 0, x * CELL_SIZE, height * CELL_SIZE);
		}

		for (int y = 0; y <= height; y++) {
			g.drawLine(0, y * CELL_SIZE, width * CELL_SIZE, y * CELL_SIZE);
		}

		g.setColor(Color.PINK);
		for (int x = 0; x <= width; x += 5) {
			g.drawLine(x * CELL_SIZE, 0, x * CELL_SIZE, height * CELL_SIZE);
		}

		for (int y = 0; y <= height; y += 5) {
			g.drawLine(0, y * CELL_SIZE, width * CELL_SIZE, y * CELL_SIZE);
		}

	}

	/**
	 * @see javax.swing.JComponent#printComponent(java.awt.Graphics)
	 */
	public void paintComponent(Graphics g) {

		super.paintComponent(g);

		paintGrid(g);
	}

	public Point pointToCell(Point p) {

		int x = p.x / CELL_SIZE;
		int y = p.y / CELL_SIZE;

		if (x >= 0
			&& x < myEstimator.getWidth()
			&& y >= 0
			&& y < myEstimator.getHeight()) {

			return new Point(x, y);
		}
		return null;
	}

	/**
	 * 
	 */
	protected void makeObservations() {
		myEstimator.observe(myScheduledObservations);
		myScheduledObservations.clear();
	}

}
