package bushfire.ui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.util.Observable;
import java.util.Observer;

import javax.swing.JComponent;
import javax.swing.JLabel;

import bushfire.model.Model;

/**
 * A user-interface element for representing the contents of the current model.
 * @author malcolmr
 */
public class ModelView extends JLabel implements Observer {

	private Model myModel;
	private byte myCursorFireType = Model.BURNING;

	private static final int CELL_SIZE = 10;

	public ModelView(Model model) {
		model.addObserver(this);
		myModel = model;

		MyMouseListener listener = new MyMouseListener();
		addMouseListener(listener);
		addMouseMotionListener(listener);
	}

	/**
	 * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
	 */
	public void update(Observable o, Object arg) {
		repaint();
	}

	public void paintComponent(Graphics g) {
		super.paintComponent(g);

		int width = myModel.getWidth();
		int height = myModel.getHeight();

		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				switch (myModel.getFireAt(x, y)) {
					case Model.UNBURNT :
						g.setColor(Color.GREEN);
						break;

					case Model.BURNING :
						g.setColor(Color.RED);
						break;

					case Model.BURNT :
						g.setColor(Color.BLACK);
						break;
				}

				g.fillRect(x * CELL_SIZE, y * CELL_SIZE, CELL_SIZE, CELL_SIZE);
			}
		}

		// draw gridlines

		g.setColor(Color.GRAY);
		for (int x = 0; x <= width; x++) {
			g.drawLine(x * CELL_SIZE, 0, x * CELL_SIZE, height * CELL_SIZE);
		}

		for (int y = 0; y <= height; y++) {
			g.drawLine(0, y * CELL_SIZE, width * CELL_SIZE, y * CELL_SIZE);
		}

	}

	/**
	 * @see java.awt.Component#getPreferredSize()
	 */
	public Dimension getPreferredSize() {
		return new Dimension(
			myModel.getWidth() * CELL_SIZE,
			myModel.getHeight() * CELL_SIZE);
	}


	/**
	 * @return
	 */
	public byte getCursorFireType() {
		return myCursorFireType;
	}

	/**
	 * @param i
	 */
	public void setCursorFireType(byte type) {
		myCursorFireType = type;
	}

	public class MyMouseListener
		extends MouseAdapter
		implements MouseMotionListener {

		/**
		* @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
		*/
		public void mouseClicked(MouseEvent e) {
			setFireAt(e.getPoint());
		}

		private void setFireAt(Point point) {
			int x = point.x / CELL_SIZE;
			int y = point.y / CELL_SIZE;
			
			if (x >= 0
				&& x < myModel.getWidth()
				&& y >= 0
				&& y < myModel.getHeight()) {
			
				myModel.setFireAt(x, y, myCursorFireType);
			}
		}

		/**
		 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
		 */
		public void mouseDragged(MouseEvent e) {
			Rectangle r = new Rectangle(e.getX(), e.getY(), 1, 1);
			((JComponent) e.getSource()).scrollRectToVisible(r);
			
			setFireAt(e.getPoint());
		}

		/**
		 * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
		 */
		public void mouseMoved(MouseEvent e) {
			//
		}

	}

}
